package owg.competent;

import java.nio.Buffer;
import java.nio.ByteBuffer;
import java.nio.FloatBuffer;
import java.nio.IntBuffer;
import java.nio.LongBuffer;
import java.nio.ShortBuffer;

/**<p>
 * This class defines the native functions for the Competent Audio Java Native Interface.
 *</p><p>
 *This version of the API is somewhat trimmed down, since many of the functions are easier to
 *re-implement in Java, compared to maintaining JNI bindings for them.
 *</p><p>
 *Where the native API would expect a CAMachine pointer, this API uses a <code>long</code>
 *holding the pointer address.
 *</p><p>
 *Where the native API would expect any other type of pointer (or array), this API uses an
 *appropriate {@link java.nio.Buffer}. Note the special case of a pointer to a CAMachine pointer
 *being represented as a {@link LongBuffer}.
 *</p><p>
 *<b>All {@link Buffer} objects passed to the native interface must be allocated with 
 *{@link ByteBuffer#allocateDirect(int)}.</b> This JNI binding will not function on JVM
 *implementations that do not support direct buffers.
 *</p><p>
 *Most of the functions in this class call their respective native CA API functions with no
 *abstraction whatsoever. Please refer to competent_audio.h for documentation of these functions.
 *It is strongly recommended that you use {@link CompetentAudio} instead.
 *</p>
 *It is generally not safe to use this file in conjunction with the high level machine objects.*/
public class CANatives
{
	/**The version number of the jcompetent_audio jni layer. 
	 * This number appears in the folder names for the native libraries to avoid version conflicts.*/
	public static final int JCOMPETENT_AUDIO_VERSION = 5;
	
	//ca_core
	public static native int getVersion();
	
	/**The native caInitialize function takes a dispatcher, allowing a C application to implement a custom dispatcher.
	 * This is not possible for a Java application, so this wrapper only allows the application to select between the two provided implementations:<ul>
	 * <li>true for ca_dispatcher_safe</li>
	 * <li>false for ca_dispatcher_unsafe</li>
	 * </ul>
	 * Furthermore, this wrapper allows a Java application to supply an {@link Appendable} to which debug information is written via the caPrint function. 
	 * The debug stream may be null if no such debug information is required.*/
	public static native int initialize(boolean threadSafe, Appendable debugStream);
	/**Like initialize, but sets a custom backend initialization order.*/
	public static native int initializeInOrder(
			int backend0, int backend1, int backend2, int backend3, int backend4, int backend5, int backend6,
			boolean threadSafe, Appendable debugStream);
	public static native int uninitialize();
	public static native int getBackendError(int backend);
	public static final int DEVICE_INDEX_AUTO = -1;
	public static native int getDeviceCount();
	public static native int getDefaultDevice();
	public static native String getDeviceName(int deviceIndex);
	public static native String getDeviceID(int deviceIndex);
	public static native int getDeviceBackend(int deviceIndex);
	public static native int refreshDeviceList();
	public static native int deleteMachine(long machine);
	/**The native caNewSink function allows the application to specify a custom sound model.
	 * This is not possible for a Java application, so this wrapper only allows for the application to select between the mono (1) or stereo (2) sound model.
	 * Passing any other value results in a CA_ERROR_SPEAKER_LAYOUT.*/
	public static native int newSink(LongBuffer sinkPtr, int soundModel, int maxChunkFrames);
	public static native int openSink(long sinkMachine, int requestedDeviceIndex, int requestedSampleRate);
	public static native int getSinkError(long sinkMachine);
	public static native int getSampleRate(long machine);
	public static native int getSpeakerCount(long machine);
	public static native int resetDestinationFunc(long dst);
	public static native int setGain(long machine, double fadeTime, float gain);
	public static native int setFrequency(long machine, double fadeTime, double frequency);
	//ca_mixer
	public static native int newMixer(LongBuffer mixerPtr, long dst);
	//ca_sampler
	public static native int newSampler(LongBuffer samplerPtr, long dst);
	public static native int setInterpolation(long sampler, boolean enable);
	public static native int getInterpolation(long sampler);
	public static native int play(long sampler);
	public static native int playAt(long sampler, double startPos);
	public static native int playReset(long sampler, double startPos);
	public static native int stop(long sampler);
	public static native int pause(long sampler);
	public static native int resume(long sampler);
	public static native int ensurePlaying(long sampler);
	public static native int getState(long sampler);
	public static native int setLooping(long sampler, boolean looping);
	public static native int getLooping(long sampler);
	public static native int setLoopStart(long sampler, double loopStart);
	public static native double getLoopStart(long sampler);
	/**The native caEnqueueClip function allows the application to specify a dequeue callback for cleanup.
	 * It is not possible to do this from a Java application, so instead there are two reasonable modes:<ul>
	 * <li> if copyData is true, then the sampleData buffer is copied, and the JNI layer handles its memory.</li>
	 * <li> if copyData is false, then the CA library will use the sampleData buffer without copying its data.
	 * 		The JNI layer creates a strong reference to it, which is deleted when the sample data is no longer 
	 * 		needed. This ensures the buffer is not garbage collected while it is in use, but may be garbage 
	 * 		collected like normal when it is no longer in use.
	 *		Also, <b>if the value of littleEndian does not match the native byte order, then the buffer is 
	 *		byte-swapped to match the native order.</b></li>
	 * </ul>*/
	public static native int enqueueClip(long sampler, boolean autoDequeue, boolean copyData, 
			int channelCount, double sampleRate, int sampleFormat, boolean littleEndian, int frameCount, Buffer sampleData);
	public static native int clearClipQueue(long sampler);
	public static native int getClipQueueSize(long sampler);
	public static native int getProcessedClips(long sampler);
	
	public static native int clearChannelMatrix(long sampler);
	public static native int setChannelMatrix(long sampler, double fadeTime, int channelCount, int vectorSize, FloatBuffer matrix);
	//ca_dsp
	public static native int getDSPVersion();
	public static native int setDSP(long dst, short programSize, ByteBuffer program, short stackSize, int messageSize, Buffer message);
	public static native int sendDSPMessage(long dst, int messageSize, Buffer message);
	public static native int getDSPError(long machine, ByteBuffer routine, IntBuffer registers, ByteBuffer traceSize, ShortBuffer trace);
}
