package owg.util.nativelibraryloader;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Locale;

/**Utility to select suitable native library files for the current platform.
 * It can extract libraries that are packaged on the classpath (inside jar a file) and/or
 * load them from the file system.*/
public class NativeLibraryLoader
{
	/**The system property "os.name" in lowercase*/
	public final static String OS_NAME = System.getProperty("os.name").toLowerCase(Locale.ENGLISH);
	/**The system property "os.arch" in lowercase*/
	public final static String OS_ARCH = System.getProperty("os.arch").toLowerCase(Locale.ENGLISH);
	
	/**Set of possible values for {@link #OS_ARCH} for x64 (32 bit)*/
	public static final HashSet<String> X86 = new HashSet<String>(Arrays.asList("x86", "i386", "i486", "i586", "i686"));
	/**Set of possible values for {@link #OS_ARCH} for amd64, aka x86_64 or x64*/
	public static final HashSet<String> AMD64 = new HashSet<String>(Arrays.asList("amd64", "x86_64"));
	/**Set of possible values for {@link #OS_ARCH} for 32 bit ARM (v7)*/
	public static final HashSet<String> ARM = new HashSet<String>(Arrays.asList("arm", "armeabi-v7a", "armeabi"));
	/**Set of possible values for {@link #OS_ARCH} for 64 bit ARM (v8)*/
	public static final HashSet<String> AARCH64 = new HashSet<String>(Arrays.asList("aarch64", "arm64-v8a"));
	
	/**Checks for a particular platform (operating system and cpu architecture)*/
	public static interface PlatformMatcher 
	{
		/**Checks for a particular platform (operating system and cpu architecture)
		 * @param osName The os.name system property in lowercase
		 * @param osArch The os.arch system property in lowercase
		 * @return true iff the given parameters matches the platform that this matcher is for*/
		public abstract boolean match(String osName, String osArch);
	}
	
	/**Used to declare that the library has support for a given platform*/
	public static class PlatformSupport
	{
		protected final PlatformMatcher system;
		protected final String subDir;
		protected final String[] libraryFiles;
		
		/**<p>
		 * Create a support declaration for a given platform.
		 * </p><p> 
		 * This implies that the required native library file(s) for this platform are included as follows:<ul>
		 * <li>If <code>subDir</code> is null, then the library files are in the same Java package as <code>localClass</code>.</li>
		 * <li>Otherwise, the library files are in the sub-package named <code>subDir</code> relative to the package that <code>localClass</code> is in,</li>
		 * </ul>
		 * where <code>localClass</code> is the Class object that is passed to the {@link NativeLibraryLoader} constructor.
		 * </p><p>
		 * In addition, if <code>useSubDir</code> is enabled for the loader, then <code>subDir</code> also affects where the library files are placed on the disk.
		 * </p>
		 * @param system A matcher to detect if the JVM is currently running on the supported platform
		 * @param subDir A sub-directory where files for this package are stored, or null, see above.
		 * @param libraryFiles The name(s) of the required library file(s).*/
		public PlatformSupport(PlatformMatcher system, String subDir, String... libraryFiles) {
			this.system = system;
			this.subDir = subDir;
			this.libraryFiles = libraryFiles;
		}
	}
	
	/**Happens when trying to create a {@link NativeLibraryLoader} on a platform that you have not declared support for.*/
	public static class PlatformUnsupportedException extends RuntimeException
	{
		public PlatformUnsupportedException(String osName, String osArch) {
			super("System not supported: "+osName+" -"+osArch);
		}
	}

	protected final Class<?> localClass;
	protected final File libraryDir;
	protected final boolean useSubDir;
	protected final PlatformSupport system;

	/**<p>
	 * Creates a loader for a native library, where the native library files are stored on the classpath (in a jar file).
	 * </p>
	 * @param localClass The library files are stored in the same package as this class, or in a sub-package relative to that package.
	 * This may be null if the jar files do not need to be extracted from the classpath.
	 * @param libraryDir The directory on the disk where the library files will be extracted to and/or loaded from.
	 * @param useSubDir If true, then the <code>subDir</code> will be created inside the <code>libraryDir</code>, 
	 * where <code>subDir</code> is the value passed to {@link PlatformSupport#PlatformSupport(PlatformMatcher, String, String...)} 
	 * for the current platform. The library files will be extracted into and/or loaded from this subdirectory. 
	 * If useSubDir is false, then files will be placed directly into the <code>libraryDir</code>.
	 * Note that if <code>subDir</code> is null, then this has no effect.
	 * @param systems Declares all the platforms supported by your native library, and the library files required on those platforms.
	 * 
	 * @throws PlatformUnsupportedException If the current platform is not supported.*/
	public NativeLibraryLoader(Class<?> localClass, File libraryDir, boolean useSubDir, PlatformSupport... systems) throws IllegalArgumentException, PlatformUnsupportedException
	{
		this.localClass = localClass;
		this.useSubDir = useSubDir;
		for(PlatformSupport system : systems)
		{
			if(system.system.match(OS_NAME, OS_ARCH))
			{
				this.system = system;
				if(useSubDir && system.subDir != null)
					this.libraryDir = new File(libraryDir, system.subDir);
				else
					this.libraryDir = libraryDir;
				return;
			}
		}
		throw new PlatformUnsupportedException(OS_NAME, OS_ARCH);
	}
	
	/**<p>
	 * Extracts the library files for the current platform to the library directory,
	 * as defined by {@link #NativeLibraryLoader(Class, File, boolean, PlatformSupport...)}.
	 * </p>
	 * @throws IOException If there was an error extracting the library.
	 * Note that if a library file already exists and cannot be deleted, 
	 * this is not considered an error. The contents of such files are
	 * however not checked by this method.
	 * @throws IllegalStateException If <code>localClass</code> was passed as null to the constructor.*/
	public void extract() throws IOException, IllegalStateException 
	{
		if(localClass == null)
			throw new IllegalArgumentException("Cannot extract when localClass is null.");
		libraryDir.mkdirs();
		if(!libraryDir.isDirectory())
			throw new IOException("The directory does not exist, and cannot be created: "+libraryDir);
		
		for(String fname : system.libraryFiles)
		{
			//Always check that the resource exists
			String resourceName = (system.subDir!=null?system.subDir+'/'+fname:fname);
			URL resource = localClass.getResource(resourceName);
			if(resource == null)
				throw new IOException("Resource not found: localClass="+localClass.getName()+", resourceName="+resourceName);
			
			File f = new File(libraryDir, fname);
			//Try to remove the file if it's already there
			f.delete();
			//Don't care if it fails
			if(f.isFile())
				continue;
			
			InputStream in = null;
			FileOutputStream out = null;
			try {
				in = resource.openStream();
				out = new FileOutputStream(f);
				
				byte[] buf = new byte[8192];
				int totalRead = 0, read;
				while((read = in.read(buf)) >= 0)
				{
					totalRead += read;
					out.write(buf, 0, read);
				}
				if(totalRead == 0)
					throw new IOException("Resource was empty: localClass="+localClass.getName()+", resourceName="+resourceName);
			} finally {
				try { in.close(); } catch(IOException err) {}
				try { out.close(); } catch(IOException err) {}
			}
		}
	}
	/**<p>
	 * Attempts to load the native library file(s) from the library directory on the disk, 
	 * as defined by {@link #NativeLibraryLoader(Class, File, boolean, PlatformSupport...)}.
	 * </p><p>
	 * If you are shipping the native library files in a jar file, then this should be called after {@link #extract(File)}.
	 * </p>
	 * @param directory The directory where the files are located. 
	 * If <code>extractToSubFolder</code> is true and the <code>subDir</code> for the current platform is non-null, 
	 * then the files are actually located in a subdirectory with that name.
	 * @throws FileNotFoundException If a library file doesn't exist.
	 * @throws UnsatisfiedLinkError If a native function was missing from a library.*/
	public void load() throws FileNotFoundException, UnsatisfiedLinkError
	{
		for(String fname : system.libraryFiles)
		{
			File f = new File(libraryDir, fname);
			if(!f.exists())
				throw new FileNotFoundException("Library file does not exist: "+f);
			System.load(f.getAbsolutePath());
		}
	}

	/**Returns true if the given os.name matches any Linux version.*/
	public static boolean isLinux(String osName)
	{
		return osName.startsWith("linux");
	}
	/**Returns true if the given os.name matches any Windows version.*/
	public static boolean isWindows(String osName) 
	{
		return osName.startsWith("win");
	}
	/**Returns true if the given os.name matches any Mac OS version.*/
	public static boolean isMacOs(String osName) 
	{
		return osName.startsWith("mac os");
	}
}
